#!/usr/bin/env python

import subprocess
import argparse
from packaging.version import Version
from pathlib import Path
import os
import shutil
import sys

from utils import APP_HOME, check_trust4_version, die, get_version


parser = argparse.ArgumentParser(add_help=False, formatter_class=argparse.ArgumentDefaultsHelpFormatter, prog='cogentip analyze')

informational = parser.add_argument_group('informational arguments')
informational.add_argument('-h', '--help',
                    action='help',
                    help='show this help message and exit')
informational.add_argument('-V', '--version',
                    action='version',
                    version=f'Cogent NGS Immune Profiler v{get_version()} (%(prog)s)')

required = parser.add_argument_group('required arguments')
required.add_argument('-r', '--receptor_kit',
                    default=argparse.SUPPRESS,
                    dest='receptor_kit',
                    help='the receptor kit',
                    choices=['TCRv1', 'TCRv2', 'BCRv1','BCRv2'],
                    required=True)
required.add_argument('-s', '--species',
                    default=argparse.SUPPRESS,
                    dest='species',
                    help='the species of the sample',
                    choices=['human', 'mouse'],
                    required=True)
required.add_argument('-m', '--meta_file',
                    default=argparse.SUPPRESS,
                    dest='meta_file',
                    help='path to the file containing sample IDs and corresponding FASTQ pairs',
                    type=Path,
                    required=True)
required.add_argument('-o', '--outdir',
                    default=argparse.SUPPRESS,
                    dest='outdir',
                    help='path to the output directory',
                    type=Path,
                    required=True)

optional = parser.add_argument_group('optional arguments')
optional.add_argument('-l', '--linker_correction',
                    dest='linker_correction',
                    help='remove reads that do not have the linker sequence in the expected location',
                    action='store_true')
optional.add_argument('-u', '--umi_cutoff',
                    default=1,
                    dest='umi_cutoff',
                    help='UMI group size cutoff',
                    type=int,
                    required=False)
optional.add_argument('--ignore_umis',
                    action='store_true',
                    dest='ignore_umis',
                    help='ignore UMIs during analysis of UMI-based data')

performance = parser.add_argument_group('performance arguments')
performance.add_argument('--trust4',
                    default='1.1.4',
                    dest='trust4_version',
                    help='the TRUST4 version to use in X.Y.Z format (minimum: 1.1.4)',
                    type=str,
                    required=False)
performance.add_argument('--threads',
                    default=4,
                    dest='threads',
                    help='the number of threads to use for FASTQ Extraction and Contig Annotation',
                    type=int,
                    required=False)
performance.add_argument('-c', '--chunks',
                    default=4,
                    dest='chunks',
                    help='the number of chunks to split FASTQ files into for assembly',
                    type=int,
                    required=False)
performance.add_argument('-k', '--keep_intermediate',
                    dest='keep_intermediate',
                    help='keep some/all intermediate files (Increases run time and storage requirements)',
                    choices=[0, 1, 2, 3, 4],
                    default=0,
                    type=int,
                    required=False)
performance.add_argument('--resume',
                    action='store_true',
                    dest='resume',
                    help='resume analysis after error')

parser.add_argument('--overwrite',
                    action='store_true',
                    dest='overwrite',
                    help=argparse.SUPPRESS)
parser.add_argument('--repseq',
                    action='store_true',
                    dest='repseq',
                    help=argparse.SUPPRESS)
parser.add_argument('--skipMateExtension',
                    action='store_true',
                    dest='skipMateExtension',
                    help=argparse.SUPPRESS)
parser.add_argument('--with_dag',
                    action='store_true',
                    dest='with_dag',
                    help=argparse.SUPPRESS)
parser.add_argument('--with_report',
                    action='store_true',
                    dest='with_report',
                    help=argparse.SUPPRESS)
parser.add_argument('--with_timeline',
                    action='store_true',
                    dest='with_timeline',
                    help=argparse.SUPPRESS)
parser.add_argument('--with_trace',
                    action='store_true',
                    dest='with_trace',
                    help=argparse.SUPPRESS)

args = parser.parse_args()

check_trust4_version(args.trust4_version)

if os.path.exists(args.outdir) and args.overwrite:
    print(f'\nDelete \'{args.outdir}\' and all of its contents before running a new analysis? [Y/N]\n> ', end='')
    answer = input()
    if answer.lower() == 'y':
        shutil.rmtree(args.outdir)

if os.path.exists(args.outdir) and not args.resume:
    die(f"""Output directory already exists: {args.outdir}
Please re-run '{sys.argv[0]}' after doing one of the following:
  • change '--outdir'
  • delete the selected output directory
  • add the '--resume' parameter to complete an interrupted analysis
  • add the '--overwrite' parameter to overwrite the previous analysis in the selected output directory
""")

command = [
  "nextflow", "run", APP_HOME / "main.nf",
  "--species", args.species,
  "--meta_file", args.meta_file,
  "--receptor_kit", args.receptor_kit,
  "--outdir", args.outdir,
  "--threads", str(args.threads),
  "--chunks", str(args.chunks),
  "--umi_cutoff", str(args.umi_cutoff),
  "--use_umis", str(not args.ignore_umis),
  "--linker_correction", str(args.linker_correction),
  "--keep_intermediate", str(args.keep_intermediate),
  "--trust4_version", str(args.trust4_version),
  "--platform", sys.platform,
  "--repseq", str(args.repseq),
  "--skipMateExtension", str(args.skipMateExtension),
]

if args.resume:
  command.append("-resume")

if args.with_dag:
  command.append("-with-dag")

if args.with_report:
  command.append("-with-report")

if args.with_timeline:
  command.append("-with-timeline")

if args.with_trace:
  command.append("-with-trace")

proc = subprocess.run(command)
sys.exit(proc.returncode)
