#!/usr/bin/env python

import subprocess
import argparse
from packaging.version import Version
from pathlib import Path
import os
import sys

from utils import APP_HOME, check_trust4_version, die, get_version


parser = argparse.ArgumentParser(add_help=False, prog='cogentip report')

informational = parser.add_argument_group('informational arguments')
informational.add_argument('-h', '--help',
                    action='help',
                    help='show this help message and exit')
informational.add_argument('-V', '--version',
                    action='version',
                    version=f'Cogent NGS Immune Profiler v{get_version()} (%(prog)s)')

required = parser.add_argument_group('required arguments')
required.add_argument('-a', '--analysis_dir',
                    dest='analysis_dir',
                    help='path to the top-level directory of a previous CogentIPv2 analysis',
                    type=Path,
                    required=True)
required.add_argument('-u', '--umi_cutoffs',
                    dest='umi_cutoffs',
                    help='path to the UMI cutoffs CSV file',
                    type=Path,
                    required=True)
required.add_argument('-s', '--species',
                    dest='species',
                    help='the species of the sample',
                    choices=['human', 'mouse'],
                    required=True)

performance = parser.add_argument_group('performance arguments')
performance.add_argument('--trust4',
                    default='1.1.4',
                    dest='trust4_version',
                    help='the TRUST4 version to use in X.Y.Z format (minimum: 1.1.4)',
                    type=str,
                    required=False)
performance.add_argument('--resume',
                    action='store_true',
                    dest='resume',
                    help='resume analysis after error')

parser.add_argument('--with_dag',
                    action='store_true',
                    dest='with_dag',
                    help=argparse.SUPPRESS)
parser.add_argument('--with_report',
                    action='store_true',
                    dest='with_report',
                    help=argparse.SUPPRESS)
parser.add_argument('--with_timeline',
                    action='store_true',
                    dest='with_timeline',
                    help=argparse.SUPPRESS)
parser.add_argument('--with_trace',
                    action='store_true',
                    dest='with_trace',
                    help=argparse.SUPPRESS)

args = parser.parse_args()

check_trust4_version(args.trust4_version)

if not os.path.exists(args.analysis_dir):
    die(f'Analysis directory does not exist: {args.analysis_dir}')

command = [
  "nextflow", "run", APP_HOME / "main_umi_reports.nf",
  "--analysisDir", args.analysis_dir,
  "--umiCutoffsFile", args.umi_cutoffs,
  "--species", args.species,
  "--trust4_version", str(args.trust4_version),
  "--platform", sys.platform,
]

if args.resume:
  command.append("-resume")

if args.with_dag:
  command.append("-with-dag")

if args.with_report:
  command.append("-with-report")

if args.with_timeline:
  command.append("-with-timeline")

if args.with_trace:
  command.append("-with-trace")

proc = subprocess.run(command)
sys.exit(proc.returncode)
