import argparse
import csv
import os
import sys
from pathlib import Path

def get_args():
  parser = argparse.ArgumentParser()
  parser.add_argument('INPUT_FILE',
    help='path to the input file',
    nargs="?",
    default="-")
  parser.add_argument('--outdir',
    dest='OUTDIR',
    help='path to the output directory',
    required = True,
    type=Path)
  parser.add_argument('--species',
    dest='SPECIES',
    required=True)
  parser.add_argument('--sampleID',
    dest='SAMPLEID',
    help='sampleID of the file',
    required=True)
  parser.add_argument('--isotype',
    dest='ISOTYPE',
    help='isotype of the file',
    required=True)
  parser.add_argument('--umi_cutoff',
    dest='UMI_CUTOFF',
    help='Custom UMI cutoff',
    required=False)
  return parser.parse_args()

def process_file(fh, species, sampleID, isotype, outdir, umi_cutoff=None):
  reader = csv.DictReader(fh, delimiter='\t')
  fieldnames = ["organism_id","sample_processing_id","read_count","fraction","clonal_sequence","clonal_sequence_quality","cdr3_min_quality","cdr3_sequence","cdr3_amino_acid_sequence","clonal_type","frame_shift","stop_codon","junction_length_aa","v_segment","all_v_hits","d_segment","all_d_hits","j_segment","all_j_hits","c_segment","all_c_hits"]

  if umi_cutoff is None:
    umi_cutoff_description = ""
  else:
    umi_cutoff_description = f".UMI_{umi_cutoff}"

  unfiltered_basename = f"{sampleID}_{isotype}{umi_cutoff_description}.immune_viewer_report.csv"
  filtered_basename = f"{sampleID}_{isotype}{umi_cutoff_description}.immune_viewer_filtered.csv"

  unfiltered_path = outdir / unfiltered_basename
  filtered_path = outdir / filtered_basename

  with open(unfiltered_path, "w") as unfiltered_fh:
    unfiltered_writer = csv.DictWriter(unfiltered_fh, fieldnames=fieldnames)
    unfiltered_writer.writeheader()
    with open(filtered_path, "w") as filtered_fh:
      filtered_writer = csv.DictWriter(filtered_fh, fieldnames=fieldnames)
      filtered_writer.writeheader()
      for row in reader:
        if row['v_call'] != '' and row['j_call'] != '':
          writer = unfiltered_writer
        else:
          writer = filtered_writer
        if row['junction_aa'] == '':
          stop_codon = 'Stop Codon'
        else:
          stop_codon = ''
        row = {
          "organism_id": species,
          "sample_processing_id": sampleID,
          "read_count": row['consensus_count'],
          # "fraction": row['frequency'],
          "fraction": None,    # need to calc (read in input file first to calc total)
          "clonal_sequence": row['sequence_alignment'],
          "clonal_sequence_quality": None,
          "cdr3_min_quality": None,
          "cdr3_sequence": row['junction'],
          "cdr3_amino_acid_sequence": row['junction_aa'],
          "clonal_type": isotype,
          "frame_shift": None,
          "stop_codon": stop_codon,
          "junction_length_aa": len(row['junction_aa']),
          "v_segment": row['v_call'].split("*")[0],
          "all_v_hits": None,
          "d_segment": row['d_call'].split("*")[0],
          "all_d_hits": None,
          "j_segment": row['j_call'].split("*")[0],
          "all_j_hits": None,
          "c_segment": row['c_call'].split("*")[0],
          "all_c_hits": None,
        }
        writer.writerow({k: None if v == "." else v for k,v in row.items()})

def process_file_at_path(path, species, sampleID, isotype, outdir, umi_cutoff=None):
  f = lambda fh: process_file(fh, species, sampleID, isotype, outdir, umi_cutoff)
  if path == "-":
    return f(sys.stdin)
  else:
    with open(path, encoding = "ISO-8859-1") as fh:
      return f(fh)

def main(args):
  process_file_at_path(args.INPUT_FILE, args.SPECIES, args.SAMPLEID, args.ISOTYPE, args.OUTDIR, args.UMI_CUTOFF)

if __name__ == '__main__':
  main(get_args())