#!/usr/bin/env python3
# Takara Bio Inc.
# Author: Sean Sun

script_name = 'Cogent_NGS_preprocess'
version = 'CogentIP v2.0.6'
desc = """
       preprocess.py:
       A script to analyze sequence data stored in fastq files and generated from Takara Human TCR/BCR kit (with UMI).
       User options are designed to simplify analysis procedure derived from Takara protocols.
       """

# ---------- import modules ---------- #

import argparse
import csv
import logging
import re
from collections import OrderedDict
import datetime
from contextlib import ExitStack
import os
import sys
import gzip
import time
from itertools import chain
import matplotlib.pyplot as plt
import pandas as pd
from Bio import Align # Bio.Align.PairwiseAligner instead of pairwise2


# ---------- fxn | check directory, file, get time & set timer ------- #
def get_time():
    ts = time.time()
    cur_time = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')
    return cur_time

# ---------- fxn | load meta data ---------- #
def load_meta(meta_file):
    meta_dict = OrderedDict()
    try:
        infile = open(meta_file, 'r')
        next(infile)
        for line in infile:
            words = [word.strip() for word in line.rstrip().split(',')]
            sname = ''
            r1_fq = ''
            r2_fq = ''
            if len(words) == 3:
                sname, r1_fq, r2_fq = words[0], words[1], words[2]
            elif len(words) < 3:
                print( get_time() + '[ERROR] sample ' + sname + ' does not have paired-fastqs')
            if sname in meta_dict:
                print(get_time() + '[ERROR] Found duplicate sample ID: ' + sname +
                      ', please rename it in meta data file ' + meta_file, flush=True)
                sys.exit(1)
            elif '_' in sname:
                print(get_time() + '[ERROR] Found _ in sample ID: ' + sname +
                      ', please rename it in meta data file ' + meta_file, flush=True)
                sys.exit(1)
            elif len(sname) > 20:
                print(get_time() + '[ERROR] There are more than 20 characters in sample ID: '
                      + sname + ', please rename', flush=True)
                sys.exit(1)
            meta_dict[sname] = [r1_fq, r2_fq]
            for i in range(len(meta_dict[sname])):
                fq = meta_dict[sname][i]
        infile.close()
        if '' in meta_dict.keys():
            del meta_dict['']
            print(get_time() + ' [WARN] Delete empty sample ID in meta file: ' + meta_file, flush=True)
            print(get_time() + ' [WARN] After deletion, ' + str(len(meta_dict.keys())) + ' samples to process',
                  flush=True)
            print(get_time() + ' [WARN] Please check and modify meta file if the sample number is incorrect',
                  flush=True)
            print(get_time() + ' [WARN] Check if there is any extra empty line in meta file', flush=True)
        return meta_dict
    except EnvironmentError:
        print(get_time() + ' [ERROR] Unable to open meta file: ' + meta_file, flush=True)
        sys.exit(1)

def load_bcs(bc_file):
    bcs = []
    with open(bc_file) as f:
        for line in f:
            l = line.rstrip().split(',')[0]
            bcs.append(l)
    return bcs

def load_pseq(ps_file):
    with open(ps_file) as f:
        ps_dict = OrderedDict()
        for line in f:
            (cur_seq, ori_seq) = re.split(',', line.strip())
            ps_dict[cur_seq] = ori_seq
    return ps_dict

def update_meta(bcs, meta_dict, nchain):
    idx = 0
    for k in meta_dict.keys():
        meta_dict[k].append(bcs[idx:(idx+nchain)])
        idx += nchain
    return

def load_fastq(f):
    if f.endswith('.gz'):
        f = gzip.open(f, mode = 'rt')
    else:
        f = open(f)
    return(f)

def get_sample_bcs(sample, meta_dict):
    bcs = meta_dict[sample][2]
    return bcs

def get_errorbs_bytwoseqs(seq1, seq2, errbases):
#pre: seq and seq2 have the same length
    errbcount = 0
    j=0
    seqL = len(seq1)
    while errbcount <= errbases and j <seqL:
        if seq1[j]!= seq2[j]:
            errbcount += 1
        j += 1
    if errbcount <= errbases:
        return seq2, errbcount
    return '', errbcount

def get_alia_seqs(seq):
#pred: seq contains character 'R' or 'S' or 'I' or 'Y' or 'K' or 'M'
#post: resturn seqs list including all possible alias
    seqs = []
    seqs.append(seq)
    while len(seqs)>0 and (seqs[0].find('R')>=0 or seqs[0].find('S')>=0 or seqs[0].find('I')>=0 or seqs[0].find('Y')>=0 or seqs[0].find('K')>=0 or seqs[0].find('M')>=0):
        del_num = len(seqs)
        for i in range(len(seqs)):
            if seqs[i].find('R') >= 0:
                myseq = seqs[i].replace('R', 'A', 1)
                seqs.append(myseq)
                myseq = seqs[i].replace('R', 'G', 1)
                seqs.append(myseq)
            elif seqs[i].find('S') >= 0:
                myseq = seqs[i].replace('S', 'C', 1)
                seqs.append(myseq)
                myseq = seqs[i].replace('S', 'G', 1)
                seqs.append(myseq)
            elif seqs[i].find('I') >= 0:
                myseq = seqs[i].replace('I', 'A', 1)
                seqs.append(myseq)
                myseq = seqs[i].replace('I', 'C', 1)
                seqs.append(myseq)
                myseq = seqs[i].replace('I', 'T', 1)
                seqs.append(myseq)
            elif seqs[i].find('Y') >= 0:
                myseq = seqs[i].replace('Y', 'C', 1)
                seqs.append(myseq)
                myseq = seqs[i].replace('Y', 'T', 1)
                seqs.append(myseq)
            elif seqs[i].find('K') >= 0:
                myseq = seqs[i].replace('K', 'G', 1)
                seqs.append(myseq)
                myseq = seqs[i].replace('K', 'T', 1)
                seqs.append(myseq)
            elif seqs[i].find('M') >= 0:
                myseq = seqs[i].replace('M', 'A', 1)
                seqs.append(myseq)
                myseq = seqs[i].replace('M', 'C', 1)
                seqs.append(myseq)
        if len(seqs) > del_num:
            del seqs[0:del_num]
    return seqs

# pre: prs[0] and bc should be same length
def ham_check_flex(prs, bc):
    errbs = int(len(bc)//4)
    seqs = list(prs.keys())
    for i in range(len(seqs)):
        # no insert/delete
        seq = seqs[i]
        mappedseq, mappederr = get_errorbs_bytwoseqs(bc, seq, errbs)
        if len(mappedseq)>0:
            return mappedseq
        if seq.find('R') >= 0 or seq.find('S') >= 0 or seq.find('I') >= 0 or seq.find('Y') >= 0 or seq.find('K') >= 0 or seq.find('M') >= 0:
            seqalias = get_alia_seqs(seq)
            for i in range(len(seqalias)):
                mappedseq, mappederr = get_errorbs_bytwoseqs(bc, seqalias[i], errbs)
                if len(mappedseq)>0:
                    return seq

    return ''

def get_dyn_align_score(bc, seq, match, mismatch, opengap, extgap):
#pre: the function to replace pairwise2.align.globalms(bc, seq, match, mismatch, opengap, extgap)[0][2]
#post: return alignment score
    #score = pairwise2.align.globalms(bc, seq, match, mismatch, opengap, extgap)[0][2]
    #alignments = aligner.align(bc, seq)
    maxscore = aligner.align(bc, seq)[0].score
    """
    maxscore = 0
    for alignment in sorted(alignments):
        print("Score = %.1f:" % alignment.score)
        print(alignment)
        if alignment.score > maxscore:
            maxscore = alignment.score
    if debug and maxscore > len(bc):
        print(" bc=%s\nseq=%s\noldscore=%f\nmaxscore=%f" % (bc, seq, score, maxscore) )
    """
    return maxscore-2, maxscore

def best_dynamic_score(bc, seq, match, mismatch, opengap, extgap):
    score = -1
    seqs = get_alia_seqs(seq)
    for count in range(len(seqs)):
        myseq=seqs[count]
        #myscore = pairwise2.align.globalms(bc, myseq, match, mismatch, opengap, extgap)[0][2]
        myscore, maxscore = get_dyn_align_score(bc, myseq, match, mismatch, opengap, extgap)
        #if debug and maxscore > len(bc):
        #    print(" bc=%s\nseq=%s\noldscore=%f\nmaxscore=%f" % (bc, seq, score, maxscore) )
        if myscore > len(bc):
            return myscore
        elif myscore > score:
            score = myscore
    return score

def save_to_pass_ambiguous_chain_map(bc, seq, score, p_args, sample):
    if p_args.internal_test and score == len(bc)+1:
        fname = 'pass_ambiguous_chain_map_' + sample + '.csv'
        f = open(fname, 'a')
        if not os.path.exists(fname) or os.stat(fname).st_size == 0:
            f.write("seq,chain\n")
        f.write(bc + ',' + seq + '(' + prs[seq] + ')\n')
        f.close()

def align_prs_bestscore(sample, prs, bc):
    match=2
    mismatch=-2
    opengap=-2
    extgap=-1
    if bc.find('GGGGGG') >= 0:
        return ''
    seqs = list(prs.keys())
    best_score = 0
    best_id = -1
    for i in range(len(seqs)):
        seq = seqs[i]
        score, maxscore = get_dyn_align_score(bc, seq, match, mismatch, opengap, extgap)
        if score > len(bc):
            save_to_pass_ambiguous_chain_map(bc, seq, score, p_args, sample)
            return seq
        elif seq.find('R')>=0 or seq.find('S')>=0 or seq.find('I')>=0 or seq.find('Y')>=0 or seq.find('K')>=0 or seq.find('M')>=0:
            score = best_dynamic_score(bc, seq, match, mismatch, opengap, extgap)
            if score > len(bc):
                save_to_pass_ambiguous_chain_map(bc, seq, score, p_args, sample)
                return seq
        if score > best_score:
            best_score = score
            best_id = i
    if p_args.internal_test and best_score == len(bc):
        fname = 'fail_ambiguous_chain_map_' + sample + '.csv'
        f = open(fname, 'a')
        if not os.path.exists(fname) or os.stat(fname).st_size == 0:
            f.write("seq,chain\n")
        f.write(bc + ',' + seqs[best_id] + '(' + prs[seqs[best_id]] + ')\n')
        f.close()
    return ''

"""
def align_prs_bestscore(sample, prs, bc):
    match=2
    mismatch=-2
    opengap=-2
    extgap=-1
    if bc.find('GGGGGG') >= 0:
        return ''
    seqs = list(prs.keys())
    best_score = 0
    best_id = -1
    for i in range(len(seqs)):
        seq = seqs[i]
        # match: 2; miss matched: -1; init gap: -1; cont. gap: -0.5
        #score = pairwise2.align.globalms(bc, seqs[i], 2, -1, -1, -.5)[0][2]
        score = pairwise2.align.globalms(bc, seq, match, mismatch, opengap, extgap)[0][2]
        #alignments = aligner.align(bc, seq)
        #for alignment in sorted(alignments):
        #    print("Score = %.1f:" % alignment.score)
        #    print(alignment)
        #test_alignments = pairwise2.align.localds(seq1, seq2, blosum62, -10, -1)
        if seq.find('R')>=0 or seq.find('S')>=0 or seq.find('I')>=0 or seq.find('Y')>=0 or seq.find('K')>=0 or seq.find('M')>=0:
            score = best_dynamic_score(bc, seq, match, mismatch, opengap, extgap)
        if score > best_score:
            best_score = score
            best_id = i
    if best_score > len(bc):
        if p_args.internal_test and best_score == len(bc)+1:
            fname = os.path.join(p_args.out_dir, 'pass_ambiguous_chain_map_' + sample + '.csv')
            f = open(fname, 'a')
            if not os.path.exists(fname) or os.stat(fname).st_size == 0:
                f.write("seq,chain\n")
            f.write(bc + ',' + seqs[best_id] + '(' + prs[seqs[best_id]] + ')\n')
            f.close()
        return seqs[best_id]
    elif p_args.internal_test and best_score == len(bc):
        fname = os.path.join(p_args.out_dir, 'fail_ambiguous_chain_map_' + sample + '.csv')
        f = open(fname, 'a')
        if not os.path.exists(fname) or os.stat(fname).st_size == 0:
            f.write("seq,chain\n")
        f.write(bc + ',' + seqs[best_id] + '(' + prs[seqs[best_id]] + ')\n')
        f.close()
    #else: print(bc)
    return ''
"""

def get_partial_primer_start_end(p_args, prs):
    if p_args.receptor_kit == 'TCRv1':
        start, end = 18, 24
        if p_args.species == 'mouse':
            start, end = 7, 13
    elif p_args.receptor_kit == 'TCRv2':
        start, end = 5, 11
        if p_args.species == 'mouse':
            start, end = 7, 13
    elif p_args.receptor_kit == 'BCRv1':
        start, end = 16, 22
        if p_args.species == 'mouse':
            start, end = 16, 23
    elif p_args.receptor_kit == 'BCRv2' or p_args.receptor_kit == 'BCRv2sub' or p_args.receptor_kit == "BCRv2hy":
        start, end = 17, 24
        if p_args.species == 'human' and len(list(prs.keys())[0])>9:
            start, end = 17-3, 26
        if p_args.species == 'mouse':
            start, end = 16-3, 27

    if end-start != len(list(prs.keys())[0]): # mouse BCRv1
        #print("check and adjust the length of bc")
        if len(list(prs.keys())[0])>11 and p_args.receptor_kit == 'BCRv1' and p_args.species == 'mouse':
            start = start - 2
        end = start + len(list(prs.keys())[0])
    return start, end

def get_subseq(read, repo_type, species, pseq_dict, prs, len_cutoff=30, mismatch=1):
    start, end = get_partial_primer_start_end(p_args, prs)
    if len(read[1]) >= len_cutoff:
        subseq = read[1][start: end]
    else:
        subseq = 'short'
    if mismatch == 0 or subseq in prs.keys() or subseq == 'short':
        pass
    else:
        #rawseq = ham_check(pseq_dict, subseq)
        if p_args.species == 'mouse' and p_args.receptor_kit.find('BCR')>=0:
            rawseq = ham_check_flex(prs, subseq)
            if len(rawseq) == 0:
                rawseq = align_prs_bestscore(sample, prs, subseq)
        else:
            rawseq = ham_check_flex(prs, subseq)
        if (len(rawseq) > 0):
            subseq = rawseq
    return subseq

def get_umi(read, umi_start=0, len_umi=12):
    if len(read[1]) < len_umi:
        umi = 'na'
    else:
        umi = read[1][umi_start : umi_start + len_umi]
    return umi

def process_read(r1, r2, bcs, bc_idx, umi, p_args):
    # This function modifies r1 and r2 as out parameters
    if (p_args.receptor_kit in ['TCRv2', 'BCRv2'] or (p_args.receptor_kit == "BCRv1" and p_args.species == "human")) and p_args.library_kit == 'takara_smartseq':
        bc = bcs[bc_idx].replace('"', '')
        tmp = r1[0].split()
        tmp[0] += '_bc_' + bc + '_umi_' + umi
        r1[0] = ' '.join(tmp)

        tmp = r2[0].split()
        tmp[0] += '_bc_' + bc + '_umi_' + umi
        r2[0] = ' '.join(tmp)
        r2[1] = bc + r2[1]
        r2[3] = 'I' * len(bc) + r2[3]

def correct_linker(read, start=12, linker='GTAC'):
    seq = read[1][start:start + len(linker)]
    if seq != linker:
        return True
    else:
        return False

def demux_subnoumi(sample, p_args, meta_dict, prs, sd):
    bcs = get_sample_bcs(sample, meta_dict)
    r1 = []
    r2 = []
    raw_r1_fq = meta_dict[sample][0]
    raw_r2_fq = meta_dict[sample][1]
    out_r1_fqs = []
    out_r2_fqs = []
    chaintypes = []  # subseqs + short  + undetermined + flc
    seqtypes = []  # subseqs + short  + undetermined + flc
    tmpcts = [0] * (len(prs) + 3)  # subseqs + short  + undetermined + flc
    for key2, val2 in prs.items():
        subseq = key2
        chain = val2
        name = sample + '_' + chain
        out_r1_fq = name + '_R1.fastq'
        out_r2_fq = name + '_R2.fastq'
        out_r1_fqs.append(out_r1_fq)
        out_r2_fqs.append(out_r2_fq)
        seqtypes.append(subseq)
        bc_idx = seqtypes.index(subseq)
        # print('bc_idx=',bc_idx)
        # print('bc=',bcs[bc_idx])
        bc = bcs[bc_idx].replace('"', '')
        chaintypes.append(chain + '_' + bc)
    short_r1_fq = sample + '_short_R1.fastq'
    short_r2_fq = sample + '_short_R2.fastq'
    out_r1_fqs.append(short_r1_fq)
    out_r2_fqs.append(short_r2_fq)
    seqtypes.append('short')
    chaintypes.append('short')
    unde_r1_fq = sample + '_undetermined_R1.fastq'
    unde_r2_fq = sample + '_undetermined_R2.fastq'
    out_r1_fqs.append(unde_r1_fq)
    out_r2_fqs.append(unde_r2_fq)
    seqtypes.append('undetermined')
    chaintypes.append('undetermined')
    if p_args.linker_correction:
        flc_r1_fq = sample + '_flc_R1.fastq'
        flc_r2_fq = sample + '_flc_R2.fastq'
        out_r1_fqs.append(flc_r1_fq)
        out_r2_fqs.append(flc_r2_fq)
        seqtypes.append('flc')
        chaintypes.append('flc')
    else:
        chaintypes.append('flc')

    for item in out_r1_fqs:
        open(item, 'w+').close()
    for item in out_r2_fqs:
        open(item, 'w+').close()
    fhs1 = []
    fhs2 = []
    with load_fastq(raw_r1_fq) as f1, load_fastq(raw_r2_fq) as f2, ExitStack() as cm:
        for name in out_r1_fqs:
            fhs1.append(cm.enter_context(open(name, 'a')))
        for name in out_r2_fqs:
            fhs2.append(cm.enter_context(open(name, 'a')))
        for x, y in zip(f1, f2):
            r1.append(x.rstrip())
            r2.append(y.rstrip())
            if len(r1) == 4:
                subseq = get_subseq(r1, p_args.receptor_kit, p_args.species,
                                    pseq_dict, prs, len_cutoff=30, mismatch=1)
                umi = ""
                if p_args.receptor_kit != "TCRv1" and (p_args.receptor_kit != "BCRv1" or p_args.species != "mouse"):
                    umi = get_umi(r2)
                if subseq == 'short':
                    tmpcts[len(prs)] += 1
                    for item in r1:
                        fhs1[seqtypes.index(subseq)].write('%s\n' % item)
                    for item in r2:
                        fhs2[seqtypes.index(subseq)].write('%s\n' % item)
                elif subseq in prs:
                    if p_args.linker_correction:
                        failed_lc = correct_linker(r2)
                        if failed_lc:
                            tmpcts[len(prs) + 2] += 1
                            for item in r1:
                                fhs1[seqtypes.index('flc')].write('%s\n' % item)
                            for item in r2:
                                fhs2[seqtypes.index('flc')].write('%s\n' % item)
                        else:
                            bc_idx = seqtypes.index(subseq)
                            process_read(r1, r2, bcs, bc_idx, umi,p_args)
                            tmpcts[bc_idx] += 1
                            for item in r1:
                                fhs1[seqtypes.index(subseq)].write('%s\n' % item)
                            for item in r2:
                                fhs2[seqtypes.index(subseq)].write('%s\n' % item)
                    else:  # no linker correction
                        bc_idx = seqtypes.index(subseq)
                        process_read(r1, r2, bcs, bc_idx, umi, p_args)
                        tmpcts[bc_idx] += 1
                        for item in r1:
                            fhs1[seqtypes.index(subseq)].write('%s\n' % item)
                        for item in r2:
                            fhs2[seqtypes.index(subseq)].write('%s\n' % item)
                else: # undetermined
                    tmpcts[len(prs) + 1] += 1
                    for item in r1:
                        fhs1[seqtypes.index('undetermined')].write('%s\n' % item)
                    for item in r2:
                        fhs2[seqtypes.index('undetermined')].write('%s\n' % item)
                del r1[:]
                del r2[:]
    total = sum(tmpcts)
    tmpcts.append(total)
    chaintypes.append('total')
    proportion = [round(x/total, 3) for x in tmpcts]
    sd[sample] = list(zip(chaintypes, tmpcts, proportion))
    return

def get_flc_reads(sample, r1, r2, subseq, prs, umi, p_args):
    fname = 'flc_info_' + sample + '.csv'
    f = open(fname, 'a')
    if not os.path.exists(fname) or os.stat(fname).st_size == 0:
        f.write("sample,umi,linker,sim(GTACGGG),pos_GTACGGG,p0_40_sim_max,p0_40,p0_40_seq,primer_seq,partial_primer,expected_chain,r1,r2\n")
    start, end = get_partial_primer_start_end(p_args, prs)
    #sim_score = pairwise2.align.globalms(r2[1][12:19], 'GTACGGG', 1, -1, -1, -.5)[0][2]
    sim_score, temp = get_dyn_align_score(r2[1][12:19], 'GTACGGG', 2, -2, -2, -1)
    pos = r2[1].find('GTACGGG')
    if pos >= 0 and pos < 5:
        pos_next = r2[1].find('GTACGGG', pos+1)
        if pos_next >= 0 and pos_next < 30:
            pos = pos_next
    p0_40 = -1
    p0_40_sim_max = -10
    p0_40_seq = 'NNNNNNN'
    if pos >= 0 and pos <= 40:
        p0_40 = pos
        p0_40_sim_max = 7
        p0_40_seq = r2[1][pos:pos+7]
    elif pos == -1 or pos > 40:
        stop = len(r2[1])-7
        if stop > 40:
            stop = 40
        for i in range(stop):
            #s = pairwise2.align.globalms(r2[1][i:i+7], 'GTACGGG', 1, -1, -1, -.5)[0][2]
            s, temp = get_dyn_align_score(r2[1][12:19], 'GTACGGG', 2, -2, -2, -1)
            if s > p0_40_sim_max:
                p0_40_sim_max = s
                p0_40 = i
                p0_40_seq = r2[1][i:i+7]
    line = sample + ',' + umi + ','+ r2[1][12:19] + ',' + str(sim_score) + ',' + str(pos) + ',' + str(p0_40_sim_max) + ',' + str(p0_40) + ','\
         + p0_40_seq + ',' + r1[1][7:27] + ',' + r1[1][start:end] + ',(' + prs[subseq] + ')_' + subseq + ',' + r1[1] + ',' + r2[1] + '\n'
    f.write(line)
    f.close()
    return

def demux_sub(sample, p_args, meta_dict, prs, sd):
    bcs = get_sample_bcs(sample, meta_dict)
    r1 = []
    r2 = []
    raw_r1_fq = meta_dict[sample][0]
    raw_r2_fq = meta_dict[sample][1]
    out_r1_fqs = []
    out_r2_fqs = []
    chaintypes = []  # subseqs + short  + undetermined + flc
    seqtypes = []  # subseqs + short  + undetermined + flc
    tmpcts = [0] * (len(prs) + 3)  # subseqs + short  + undetermined + flc
    for key2, val2 in prs.items():
        subseq = key2
        chain = val2
        name = sample + '_' + chain
        out_r1_fq = name + '_R1.fastq'
        out_r2_fq = name + '_R2.fastq'
        out_r1_fqs.append(out_r1_fq)
        out_r2_fqs.append(out_r2_fq)
        seqtypes.append(subseq)
        bc_idx = seqtypes.index(subseq)
        # print('bc_idx=',bc_idx)
        # print('bc=',bcs[bc_idx])
        bc = bcs[bc_idx].replace('"', '')
        chaintypes.append(chain + '_' + bc)
    short_r1_fq = sample + '_short_R1.fastq'
    short_r2_fq = sample + '_short_R2.fastq'
    out_r1_fqs.append(short_r1_fq)
    out_r2_fqs.append(short_r2_fq)
    seqtypes.append('short')
    chaintypes.append('short')
    unde_r1_fq = sample + '_undetermined_R1.fastq'
    unde_r2_fq = sample + '_undetermined_R2.fastq'
    out_r1_fqs.append(unde_r1_fq)
    out_r2_fqs.append(unde_r2_fq)
    seqtypes.append('undetermined')
    chaintypes.append('undetermined')
    if p_args.linker_correction:
        flc_r1_fq = sample + '_flc_R1.fastq'
        flc_r2_fq = sample + '_flc_R2.fastq'
        out_r1_fqs.append(flc_r1_fq)
        out_r2_fqs.append(flc_r2_fq)
        seqtypes.append('flc')
        chaintypes.append('flc')
    else:
        chaintypes.append('flc')
    for item in out_r1_fqs:
        open(item, 'w+').close()
    for item in out_r2_fqs:
        open(item, 'w+').close()
    fhs1 = []
    fhs2 = []
    with load_fastq(raw_r1_fq) as f1, load_fastq(raw_r2_fq) as f2, ExitStack() as cm:
        for name in out_r1_fqs:
            fhs1.append(cm.enter_context(open(name, 'a')))
        for name in out_r2_fqs:
            fhs2.append(cm.enter_context(open(name, 'a')))
        for x, y in zip(f1, f2):
            r1.append(x.rstrip())
            r2.append(y.rstrip())
            if len(r1) == 4:
                subseq = get_subseq(r1, p_args.receptor_kit, p_args.species,
                                    pseq_dict, prs, len_cutoff=30, mismatch=1)
                umi = get_umi(r2)
                if subseq == 'short':
                    tmpcts[len(prs)] += 1
                    for item in r1:
                        fhs1[seqtypes.index(subseq)].write('%s\n' % item)
                    for item in r2:
                        fhs2[seqtypes.index(subseq)].write('%s\n' % item)
                elif subseq in prs:
                    if p_args.linker_correction:
                        failed_lc = correct_linker(r2)
                        if failed_lc:
                            tmpcts[len(prs) + 2] += 1
                            for item in r1:
                                fhs1[seqtypes.index('flc')].write('%s\n' % item)
                            for item in r2:
                                fhs2[seqtypes.index('flc')].write('%s\n' % item)
                            if p_args.internal_test:
                                get_flc_reads(sample, r1, r2, subseq, prs, umi, p_args)
                        else:
                            bc_idx = seqtypes.index(subseq)
                            process_read(r1, r2, bcs, bc_idx, umi, p_args)
                            tmpcts[bc_idx] += 1
                            for item in r1:
                                fhs1[seqtypes.index(subseq)].write('%s\n' % item)
                            for item in r2:
                                fhs2[seqtypes.index(subseq)].write('%s\n' % item)
                    else:  # no linker correction
                        bc_idx = seqtypes.index(subseq)
                        process_read(r1, r2, bcs, bc_idx, umi, p_args)
                        tmpcts[bc_idx] += 1
                        for item in r1:
                            fhs1[seqtypes.index(subseq)].write('%s\n' % item)
                        for item in r2:
                            fhs2[seqtypes.index(subseq)].write('%s\n' % item)
                else: # undetermined
                    tmpcts[len(prs) + 1] += 1
                    for item in r1:
                        fhs1[seqtypes.index('undetermined')].write('%s\n' % item)
                    for item in r2:
                        fhs2[seqtypes.index('undetermined')].write('%s\n' % item)
                del r1[:]
                del r2[:]
    total = sum(tmpcts)
    tmpcts.append(total)
    chaintypes.append('total')
    proportion = [round(x/total, 3) for x in tmpcts]
    sd[sample] = list(zip(chaintypes, tmpcts, proportion))
    return

def write_sample_qc(p_args, sd):
    with open(f'{p_args.sample}.QC_stats.csv', 'w', newline='') as qc_file:
        writer = csv.writer(qc_file, delimiter=',', lineterminator="\n")
        col_names = ['Sample_ID']
        key1 = list(sd)[0]
        for item in sd[key1]:
            col_names.append(item[0].split('_')[0])
            col_names.append(item[0].split('_')[0]+'%')
        writer.writerow(col_names)
        for key, val in sd.items():
            sample_id = key
            # stats = [ str(item[1]) + '(' + '{:.1%}'.format(item[2])+ ')' for item in val ]
            l1 = [str(item[1]) for item in val ]
            l2 = ['{:.1%}'.format(item[2]) for item in val]
            stats = list(chain.from_iterable(zip(l1, l2)))
            writer.writerow([sample_id] + stats)

def write_airr_sample_qc(p_args, sd):
    with open(f'{p_args.sample}.QC_stats.airr.csv', 'w', newline='') as qc_file:
        writer = csv.writer(qc_file, delimiter=',', lineterminator="\n")
        col_names = ['organism_id', 'sample_processing_id']
        key1 = list(sd)[0]
        for item in sd[key1]:
            col_names.append(item[0].split('_')[0])
            col_names.append(item[0].split('_')[0]+'%')
        writer.writerow(col_names)
        organism = p_args.species
        for key, val in sd.items():
            sample_id = key
            # stats = [ str(item[1]) + '(' + '{:.1%}'.format(item[2])+ ')' for item in val ]
            l1 = [str(item[1]) for item in val ]
            l2 = ['{:.1%}'.format(item[2]) for item in val]
            stats = list(chain.from_iterable(zip(l1, l2)))
            writer.writerow([organism] + [sample_id] + stats)

def create_fd(sd):
    sublist = ['short', 'undetermined', 'flc', 'total']
    pd = OrderedDict((k, sd[k]) for k in sd.keys())
    for k, v in pd.items():
        l1 = []
        l2 = []
        for i in range(len(v)):
            if v[i][1] > 0:  # cutoff to discard chain category, used to be v[i][2] >= 0.05
                l1.append(i)
            if v[i][0] not in sublist:
            #if v[i][0] not in ['short', 'undetermined', 'flc', 'total']:
                l2.append(i)
        l = [value for value in l1 if value in l2]
        pd[k] = [v[idx] for idx in l]
    fd = OrderedDict()
    for k, v in pd.items():
        sample = k
        bc = ''
        for i in range(len(v)):
            chain = v[i][0].split('_')[0]
            if chain == 'undetermined' or chain == 'flc':
                bc = chain #bc = v[0][0].split('_')[1]
            else:
                bc = v[i][0].split('_')[1]
            name = sample + '_' + chain
            fd[name] = [ name + '_R1.fastq', name + '_R2.fastq', bc]
    return fd

# plot pie for all QC chains(flc 6%)
def plot_QC_chains(fname):
    f = open(fname, 'r')
    dirname = os.path.dirname(fname)
    flist = f.readlines()
    labels = []
    for i in range (0,len(flist)):
        segs = flist[i].split(",")
        y = []
        mylabels = []
        for j in range(1, len(segs)-2):
            if i == 0:
                if j%2 == 1:
                    labels.append(segs[j])
            else:
                if j%2 == 1:
                    value = int(segs[j])
                    if value > 1:
                        y.append(value)
                if j%2 == 0 and value > 1:
                    mylabels.append(labels[j//2-1]+'_'+segs[j])
        if i > 0:
            ofname = os.path.join(os.path.dirname(dirname), 'QC_pie_plot_' + segs[0] + ".png")
            y[1], y[2] = y[2], y[1]
            mylabels[1], mylabels[2] = mylabels[2], mylabels[1]
            if len(y)>7:
                y[5], y[6] = y[6], y[5]
                mylabels[5], mylabels[6] = mylabels[6], mylabels[5]
            myexplode = [0]*len(y)
            myexplode[len(myexplode)-1]=0.2
            plt.pie(y, labels=mylabels, explode = myexplode)
            plt.title("QC plot sample_" + segs[0])
            #plt.show()
            of = plt.savefig(ofname)
            #of = plt.savefig(ofname, dpi=100)
            plt.close(of)
    return

def flc_analysis(fname):
    dirname = os.path.dirname(fname)
    count = 0
    heads = []
    data = []
    with open(fname, 'r') as fp:
        for line in fp:
            line = line.strip()
            segs = line.split(',')
            if count == 0:
                heads = segs.copy()
            else:
                data.append(segs[1:len(segs)-2])
            count += 1
    df = pd.DataFrame(data=data, columns=heads[1:len(heads)-2])
    total = len(df)
    df['pos_GTACGGG'] = df['pos_GTACGGG'].astype('int32')
    df['p0_40'] = df['p0_40'].astype('int32')
    df['sim(GTACGGG)'] = df['sim(GTACGGG)'].astype('float32')
    df['p0_40_sim_max'] = df['p0_40_sim_max'].astype('float32')

    for key in ['pos', 'sim_max']:
        y=[]
        mylabels = []
        if key == 'pos':
            df = df.sort_values(by=['pos_GTACGGG', 'p0_40_sim_max'], ascending=False)
            labels = ['p-1(0_40sim<3):','p-1(0_40sim=3):','p-1(0_40sim=4):','p-1(0_40sim=5):','p-1(0_40sim=6):','p0-10(sim=7):','p11(sim=7):','p13(sim=7):','p14-40(sim=7):','p41-117(sim=7):' ]
            for i in range(len(labels)):
                if i == 0:
                    sub_df = df.loc[(df['pos_GTACGGG'] ==-1) & (df['p0_40_sim_max'] < 3)]
                elif i == 1:
                    sub_df = df.loc[(df['pos_GTACGGG'] ==-1) & (df['p0_40_sim_max'] == 3)]
                elif i == 2:
                    sub_df = df.loc[(df['pos_GTACGGG'] ==-1) & (df['p0_40_sim_max'] == 4)]
                elif i == 3:
                    sub_df = df.loc[(df['pos_GTACGGG'] ==-1) & (df['p0_40_sim_max'] == 5)]
                elif i == 4:
                    sub_df = df.loc[(df['pos_GTACGGG'] ==-1) & (df['p0_40_sim_max'] == 6)]
                elif i == 5:
                    sub_df = df.loc[(df['pos_GTACGGG'] > -1) & (df['pos_GTACGGG'] <= 10)]
                elif i == 6:
                    sub_df = df.loc[(df['pos_GTACGGG'] == 11)]
                elif i == 7:
                    sub_df = df.loc[(df['pos_GTACGGG'] == 13)]
                elif i == 8:
                    sub_df = df.loc[(df['pos_GTACGGG'] > 13) & (df['pos_GTACGGG'] <= 40)]
                elif i == 9:
                    sub_df = df.loc[(df['pos_GTACGGG'] > 40)]
                #print(sub_df.shape[0])
                y.append(sub_df.shape[0])
                mylabels.append(labels[i]+str(sub_df.shape[0])+"("+str(int(100*len(sub_df)/total+0.5))+"%)")
        else:
            df = df.sort_values(by=['p0_40_sim_max', 'pos_GTACGGG'], ascending=False)
            labels = ['0_40sim_max<3:', '0_40sim_max=3:', '0_40sim_max=4:','0_40sim_max=5:','0_40sim_max=6:','0_40sim_max=7:' ]
            for i in range(len(labels)):
                if i == 0:
                    sub_df = df.loc[ df['p0_40_sim_max'] < 3 ]
                elif i == 1:
                    sub_df = df.loc[ df['p0_40_sim_max'] == 3 ]
                elif i == 2:
                    sub_df = df.loc[ df['p0_40_sim_max'] == 4 ]
                elif i == 3:
                    sub_df = df.loc[ df['p0_40_sim_max'] == 5 ]
                elif i == 4:
                    sub_df = df.loc[ df['p0_40_sim_max'] == 6 ]
                elif i == 5:
                    sub_df = df.loc[ df['p0_40_sim_max'] == 7 ]
                y.append(sub_df.shape[0])
                mylabels.append(labels[i]+str(sub_df.shape[0]))
        ofname = os.path.join(dirname, 'flc_pie_plot_' + key + '_' + segs[0] + ".png")
        #y[1], y[2] = y[2], y[1]
        #mylabels[1], mylabels[2] = mylabels[2], mylabels[1]
        plt.pie(y, labels=mylabels)
        plt.title("Pie plot sample_" +  key + '_' + segs[0])
        #plt.show()
        of = plt.savefig(ofname, dpi=250)
        #of = plt.savefig(ofname)
        plt.close(of)
    return

# ---------- main ---------- #
if __name__ == '__main__':

    # ---------- setup | parser ---------- #

    parser = argparse.ArgumentParser(description=desc)
    optional = parser._action_groups.pop()
    required = parser.add_argument_group('required arguments')
    parser._action_groups.append(optional)

    # ---------- setup | user options ---------- #

    # add required arguments
    required.add_argument('--sample', dest='sample',
                          help='specify sample name',
                          required=True)
    required.add_argument('-r', '--receptor_kit', dest='receptor_kit',
                          help='specify receptor kit: TCRv1, TCRv2, BCRv1 or BCRv2', choices=['TCRv1', 'TCRv2', 'BCRv1','BCRv2'],
                          required=True)
    required.add_argument('-m', '--meta_file', dest='meta_file',
                          help='a file contains sample ID and corresponding FASTQ pair',
                          required=True)

    # add optional arguments

    #optional.add_argument('-i', '--internal_test',  dest = 'internal_test',
    #                      help = 'internal test, will generate MiXCR results without ' +
    #                      'UMI-guided read collaspe [Default: False]',
    #                      action = 'store_true', default = False)
    optional.add_argument('-l', '--linker_correction', dest='linker_correction',
                          help='decide if remove reads based on sequence match of linker' +
                               ' [Default: False]',
                          action='store_true', default=False)
    optional.add_argument('-s', '--species',  dest = 'species',
                          help = 'specify the genome species: human, mouse' + ' [Default: human]',
                          default='human')
    optional.add_argument('--config_path',  dest = 'config_path',
                          help = 'specify the path to the config directory',
                          required=True)
    #optional.add_argument('-x', '--library_kit',  dest = 'library_kit',
    #                      help = 'specify library_kit: takara_smartseq, cellecta_air',
    #                      default='takara_smartseq')

    # ---------- setup | parse/check user options ---------- #
    p_args = parser.parse_args()

    if p_args.receptor_kit == 'TCRv1' or (p_args.receptor_kit == 'BCRv1' and p_args.species == 'mouse'):
        p_args.linker_correction = False

    p_args.internal_test = False #True
    p_args.library_kit = 'takara_smartseq'
    p_args.rm_adapter = False

    # ---------- setup | start ---------- #
    start_time = datetime.datetime.now()
    # ---------- setup | end ---------- #

    # --------- setup aligner | end ---------- #
    aligner = Align.PairwiseAligner()
    aligner.mode = 'global'
    aligner.match_score = 2
    aligner.mismatch_score = -2
    aligner.open_gap_score = -2
    aligner.extend_gap_score = -1
    aligner.target_end_gap_score = 0.0
    aligner.query_end_gap_score = 0.0

    # ---------- setup | load meta, create outdir ---------- #
    meta_dict = load_meta(p_args.meta_file)
    print(get_time() + ' [INFO] Loading meta data file specified by user', flush=True)

    # ---------- delete the folder only for internal_test or debug ----- #
    debug = False
    logger = logging.getLogger('temp_recoder')

    # ---------- setup | logger ---------- #
    LOG_NAME = 'preprocess'
    LOG_FILENAME = LOG_NAME + '.log'
    LOG_FORMAT = '%(levelname)s - %(asctime)s - %(name)s - %(message)s'
    LOG_FORMAT_ARG_DATE = '%Y-%m-%d %H:%M:%S'

    logger = logging.getLogger(LOG_NAME)
    logger.setLevel(logging.INFO)
    fh = logging.FileHandler(LOG_FILENAME, mode='w')
    formatter = logging.Formatter(LOG_FORMAT, datefmt=LOG_FORMAT_ARG_DATE)
    fh.setFormatter(formatter)
    logger.addHandler(fh)
    logger.info('Started script execution : ' + script_name + '_v' + version)
    logger.info('Original call : ' + ' '.join(sys.argv))

    # ----------- setup | define parameter & local resources----------- #
    umi_length = 12
    linker_sequence = 'GTAC'
    solution = []
    delta = lambda x,y,i,j: 1 if x[i] != y[j] else 0

    CONFIG_PATH = p_args.config_path

    BC_FILE = os.path.join(CONFIG_PATH, 'barcodes', 'bc_list.csv')
    if p_args.receptor_kit == 'TCRv1':
        BC_FILE = os.path.join(CONFIG_PATH, 'barcodes', 'I7_I5_index.csv')

    if p_args.receptor_kit == 'BCRv1' :
        if p_args.species == 'human':
            BCR_FILE = os.path.join(CONFIG_PATH, 'primer_maps', 'bcrv1_dict.csv')
        elif p_args.species == 'mouse':
            BCR_FILE = os.path.join(CONFIG_PATH, 'primer_maps', 'bcrv1mouse_dict.csv')
    elif p_args.receptor_kit == 'BCRv2' or p_args.receptor_kit == 'BCRv2sub' or p_args.receptor_kit == 'BCRv2hy':
        if p_args.species == 'human':
            BCR_FILE = os.path.join(CONFIG_PATH, 'primer_maps', 'bcrv2_dict.csv')
        elif p_args.species == 'mouse':
            BCR_FILE = os.path.join(CONFIG_PATH, 'primer_maps', 'bcrv2mouse_dict.csv')
    elif p_args.receptor_kit == 'TCRv2' :
        if p_args.species == 'human':
            TCR_FILE = os.path.join(CONFIG_PATH, 'primer_maps', 'tcrv2_dict.csv')
        elif p_args.species == 'mouse':
            TCR_FILE = os.path.join(CONFIG_PATH, 'primer_maps', 'tcrv2mouse_dict.csv')
    elif p_args.receptor_kit == 'TCRv1' :
        if p_args.species == 'human':
            TCR_FILE = os.path.join(CONFIG_PATH, 'primer_maps', 'tcrv1_dict.csv')
        elif p_args.species == 'mouse':
            TCR_FILE = os.path.join(CONFIG_PATH, 'primer_maps', 'tcrv1mouse_dict.csv')

    align_stats_dict = OrderedDict([('Total sequencing reads', 'total reads'),
        ('Successfully aligned reads', 'aligned reads'),
        ('Overlapped', 'overlapped'),
        ('Overlapped and aligned', 'overlapped and aligned')])
    clone_stats_dict = OrderedDict([('Final clonotype count', 'clonotype count')])
    if p_args.internal_test:
        clone_stats_dict = OrderedDict([('Average number of reads per clonotype', 'reads per clonotype'),
            ('Reads used in clonotypes, percent of total', 'Reads used in clonotypes'),
            ('Mapped low quality reads, percent of used', 'Mapped low quality reads'),
            ('Reads clustered in PCR error correction, percent of used', 'PCR error correction'),
            ('Reads dropped due to failed mapping, percent of total', 'failed mapping'),
            ('Final clonotype count', 'clonotype count')])

    # ---------- load bcs & receptor info -------- #

    bcs = load_bcs(BC_FILE)
    # print(bcs)
    if p_args.receptor_kit == 'TCRv1':
        nchain = 2
        prs_tcr = OrderedDict([('GTACAC', 'TRA'),('TCTGAT', 'TRB')])
        if p_args.species == 'mouse':
            prs_tcr = OrderedDict([('AACTGG', 'TRA'),('TTGGGT', 'TRB')])
    elif p_args.receptor_kit == 'TCRv2':
        nchain = 2
        prs_tcr = OrderedDict([('CAGGGT', 'TRA'),('GGAACA', 'TRB')])
        if p_args.species == 'mouse':
            prs_tcr = OrderedDict([('AACTGG', 'TRA'),('TTGGGT', 'TRB')])
            #prs_tcr = OrderedDict([('AACTGGTACAC', 'TRA'),('TTGGGTGGAGT', 'TRB')])
    elif p_args.receptor_kit == 'BCRv1':
        nchain = 4
        prs_bcr = OrderedDict([
                ('TGGGCC', 'IGG'),('AATTCT', 'IGM'),
                ('AAGACA', 'IGK'),('GGGAAC', 'IGL')])
        if p_args.species == 'mouse':
            nchain = 3
            #prs_bcr = OrderedDict([('AGTGGA', 'IGG'),('AAGATG', 'IGK'),('GAGGAA', 'IGL')])
            prs_bcr = OrderedDict([('AGTGGATAGAC', 'IGG'),('AAGATGGATAC', 'IGK'),('GAGGAAGGTGG', 'IGL')])
    elif p_args.receptor_kit == 'BCRv2':# or p_args.receptor_kit == 'BCRv2hy':
        nchain = 7
        prs_bcr = OrderedDict([
                ('GGGCCCT', 'IGG'),('GATGCAC', 'IGM'),
                ('AGACAGA', 'IGK'),('GGAACAG', 'IGL'),
                ('GGGAAGA', 'IGA'),('ACATCCG', 'IGD'),
                ('GAAGACG', 'IGE')])
        #prs_bcr = OrderedDict([
        #        ('GATGGGCCCTTG', 'IGG'),('GCGGATGCACTC', 'IGM'),
        #        ('TGAAGACAGATG', 'IGK'),('GYGGGAACAGAG', 'IGL'),
        #        ('AGCGGGAAGACC', 'IGA'),('AACACATCCGGA', 'IGD'),
        #        ('GGGGAAGACGGA', 'IGE')])
        if p_args.species == 'mouse':
            #prs_bcr = OrderedDict([
            #        ('AAGATGG', 'IGK'),('GGAAGGT', 'IGL'),
            #        ('GATGGGG', 'IGG'),('TTGGGAA', 'IGM'),
            #        ('TCAGTGG', 'IGA'),('TGTCAGG', 'IGD'),
            #        ('GAGGGTT', 'IGE')])
            prs_bcr = OrderedDict([
                    ('ACIGATGGGGSTGT', 'IGG'),('CATTTGGGAAGGAC', 'IGM'),
                    ('GGGAAGATGGATAC', 'IGK'),('AGRGGAAGGTGGRA', 'IGL'),
                    ('GTGTCAGTGGGTAG', 'IGA'),('ACATGTCAGGTTCC', 'IGD'),
                    ('GCTGAGGGTTCCTG', 'IGE')])
    elif p_args.receptor_kit == 'BCRv2sub':
        nchain = 2
        prs_bcr = OrderedDict([
                ('TTGACCA', 'IGG'), ('CTGGACC', 'IGA')])
    elif p_args.receptor_kit == 'BCRv2hy':
        nchain = 9
        prs_bcr = OrderedDict([
                ('GGGCCCT', 'IGG'),('GATGCAC', 'IGM'),
                ('AGACAGA', 'IGK'),('GGAACAG', 'IGL'),
                ('GGGAAGA', 'IGA'),('ACATCCG', 'IGD'),
                ('GAAGACG', 'IGE'),('TTGACCA', 'subIGG'),
                ('CTGGACC', 'subIGA')])

    update_meta(bcs, meta_dict, nchain)
    fd = OrderedDict()
    for key, vals in meta_dict.items():
        fd[key] = [vals[0], vals[1], vals[2][0] + "+" + vals[2][1]]

    if p_args.receptor_kit in ['TCRv1', 'TCRv2']:
        pseq_dict = load_pseq(TCR_FILE)
        prs = prs_tcr
        logger.info('TCR information loaded')
    elif p_args.receptor_kit in ['BCRv1', 'BCRv2', 'BCRv2sub', 'BCRv2hy']:
        pseq_dict = load_pseq(BCR_FILE)
        prs = prs_bcr
        logger.info('BCR information loaded')

    # ---------- preprocess -------- #
    if (p_args.library_kit == 'takara_smartseq'):
        logger.info('Starting preprocessing')
        print(get_time() + ' [INFO] Starting preprocessing', flush=True)
        sd = OrderedDict()
        for key, val in meta_dict.items():
            sample = key
            if p_args.receptor_kit == 'TCRv1' or (p_args.receptor_kit == 'BCRv1' and p_args.species == "mouse"):
                demux_subnoumi(sample, p_args, meta_dict, prs, sd)
            else:
                demux_sub(sample, p_args, meta_dict, prs, sd)
            if p_args.internal_test and p_args.linker_correction:
                flc_analysis("flc_info_"+sample+".csv")

        write_sample_qc(p_args, sd)
        write_airr_sample_qc(p_args, sd)
        fd = create_fd(sd)
        if p_args.internal_test:
            plot_QC_chains("sample_QC_stats.csv")
        logger.info('Completed preprocessing')
        print(get_time() + ' [INFO] Completed preprocessing', flush=True)

    # ---------- clean up & finish ---------- #
    logger.info('Analysis completed')
    print(get_time() + ' [INFO] Analysis completed', flush=True)
    sys.exit(0)
