import argparse
import csv
import os
import re
import sys
from collections import Counter
from itertools import islice
from pathlib import Path

from Bio import SeqIO


parser = argparse.ArgumentParser()
parser.add_argument('-o', '--outdir',
                    dest='outdir',
                    help='path to the output directory',
                    type=Path,
                    required=True)
parser.add_argument('-1', '--fastq1',
                    dest='fastq1',
                    help='path to "toassemble" R1 FASTQ file',
                    type=Path,
                    required=True)
parser.add_argument('-2', '--fastq2',
                    dest='fastq2',
                    help='path to "toassemble" R2 FASTQ file',
                    type=Path,
                    required=True)
parser.add_argument('-i', '--isotype',
                    dest='isotype',
                    help='The isotype of the FASTQ files being split. Used for UMI group size frequency CSV',
                    type=str,
                    required=True)
# Need to pick a proper default for -c
parser.add_argument('-c', '--chunks',
                    default=4,
                    dest='chunks',
                    help='the number of chunks to split FASTQ files into for assembly',
                    type=int,
                    required=False)

def extract_umi(record_description, p=re.compile(r'umi_(?P<umi>[ACTGN]+)')):
    m = p.search(record_description)
    return m.group('umi')

def write_umi_group_sizes_frequency_csv(umi_counts, outdir, outfile='umi_group_sizes_frequency.csv'):
    mig_sizes = Counter()
    for mig, size in umi_counts.items():
        mig_sizes[size] += 1

    with open(f'{outdir}/{outfile}', 'w', newline='') as csvfile:
        fieldnames = ['umi_group_size', 'umi_groups']
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        writer.writeheader()
        for umi_group_size, umi_groups in sorted(mig_sizes.items()):
            writer.writerow({'umi_group_size': umi_group_size, 'umi_groups': umi_groups})

args = parser.parse_args()

umi_cutoff = 1
r1_fastq_in = args.fastq1
r2_fastq_in = args.fastq2
outdir = args.outdir
n_chunks = args.chunks

os.makedirs(outdir, exist_ok=True)

chunk_barcodes = {}

umi_counts = Counter()

with open(r1_fastq_in) as r1_fastq_in_fh:
    for r1_record in SeqIO.parse(r1_fastq_in_fh, 'fastq'):
        umi = extract_umi(r1_record.description)
        umi_counts[umi] += 1

write_umi_group_sizes_frequency_csv(umi_counts, outdir, f'umi_group_sizes_frequency.{args.isotype}.csv')

if n_chunks <= 1:
    sys.exit()

umi_counts_filtered = Counter({u: c for u, c in umi_counts.items() if c >= umi_cutoff})

for i, umi_info in enumerate(umi_counts_filtered.most_common()):
    chunk_barcodes[umi_info[0]] = i % n_chunks

r1_chunk_files = [open(f'{outdir}/chunk_{i}_R1.fastq', 'w') for i in range(min(n_chunks, len(chunk_barcodes)))]
r2_chunk_files = [open(f'{outdir}/chunk_{i}_R2.fastq', 'w') for i in range(min(n_chunks, len(chunk_barcodes)))]

with open(r1_fastq_in) as r1_fastq_in_fh:
    r2_fastq_in_fh = open(r2_fastq_in)
    r2_record_generator = SeqIO.parse(r2_fastq_in_fh, 'fastq')
    for r1_record in SeqIO.parse(r1_fastq_in_fh, 'fastq'):
        r2_record = next(islice(r2_record_generator, 1))
        try:
            umi = extract_umi(r1_record.description)
            chunk_id = chunk_barcodes[umi]
        except KeyError:
            continue
        else:
            SeqIO.write(r1_record, r1_chunk_files[chunk_id], 'fastq')
            SeqIO.write(r2_record, r2_chunk_files[chunk_id], 'fastq')
